;;; -*- Mode: Lisp; Package: EDITOR -*-
;;;
;;; This file is part of xyzzy.
;;;

(provide "app-menu")

(in-package "editor")

(export '(*app-menu* *app-popup-menu* *menu-history-max*
	  *app-rectangle-popup-menu* *clipboard-char-encoding-popup-menu*
	  *dictionary-popup-menu* *paste-rectangle-popup-menu*
	  *command-bar-popup-menu*
	  make-char-encoding-popup-menu
	  add-file-history-to-menu *menu-display-length*
	  *init-app-menus-hook*))

(defvar *app-menu* nil)
(defvar *app-popup-menu* nil)
(defvar *app-rectangle-popup-menu* nil)
(defvar *clipboard-char-encoding-popup-menu* nil)
(defvar *dictionary-popup-menu* nil)
(defvar *paste-rectangle-popup-menu* nil)
(defvar *command-bar-popup-menu* nil)

(defvar *recent-file-popup-menu* nil)
(defvar *recent-session-popup-menu* nil)

(defvar *menu-history-max* 9)

(defvar *init-app-menus-hook* nil)

(defun init-app-menus ()

  (setq *clipboard-char-encoding-popup-menu*
	(make-char-encoding-popup-menu
	 #'(lambda (encoding)
	     (interactive)
	     (continue-pre-selection)
	     (setq *clipboard-char-encoding* encoding))
	 #'(lambda (encoding)
	     (and (eq *clipboard-char-encoding* encoding) :check))
	 *clipboard-char-encoding-list*))

  (setq *dictionary-popup-menu*
	(define-popup-menu
	  (:item nil "ap(&J)"
	   'lookup-j2e-dictionary-selection)
	  (:item nil "ǂ(&R)"
	   'lookup-reading-dictionary-selection)
	  (:item nil "pa(&E)"
	   'lookup-e2j-dictionary-selection)
	  (:item nil "n(&I)"
	   'lookup-idiom-dictionary-selection)
	  :sep
	  (:item nil "Ӱ(&D)"
	   'edict-mode
	   #'(lambda () (if edict-mode :check)))))

  (setq *paste-rectangle-popup-menu*
	(define-popup-menu
	  (:item nil "}(&I)"
	   'yank-rectangle
	   #'(lambda () (unless *rectangle-kill-buffer* :disable)))
	  (:item nil "㏑(&O)"
	   'overwrite-rectangle
	   #'(lambda () (unless *rectangle-kill-buffer* :disable)))
	  (:item nil "s֒ǉ(&A)"
	   'append-rectangle
	   #'(lambda () (unless *rectangle-kill-buffer* :disable)))
	  (:item nil "ɂđ}(&R)"
	   'yank-rectangle-as-region
	   #'(lambda () (unless *rectangle-kill-buffer* :disable)))))

  (setq *app-popup-menu*
	(define-popup-menu
	  (:item nil "؂(&T)"
	   'kill-selection-to-clipboard :modify-any-selection)
	  (:item nil "߰(&C)"
	   'copy-selection-to-clipboard :any-selection)
	  (:item nil "\t(&P)"
	   'paste-from-clipboard :clipboard)
	  (:item nil "\t - `(&G)"
	   'paste-rectangle-from-clipboard :clipboard)
	  (:menu nil "دް޴ݺިݸ(&E)"
	   *clipboard-char-encoding-popup-menu*)
	  (:item nil "폜(&D)"
	   'delete-selection :modify-any-selection)
	  (:item nil "SĂI(&A)"
	   'selection-whole-buffer)
	  :sep
	  (:menu nil "(&I)"
	   *dictionary-popup-menu*)
	  (:menu nil "`\t(&R)"
	   *paste-rectangle-popup-menu*)))

  (setq *app-rectangle-popup-menu*
	(define-popup-menu
	  (:item nil "؂(&T)"
	   'kill-rectangle-selection :modify-rectangle)
	  (:item nil "߰(&C)"
	   'copy-rectangle-selection :rectangle)
	  (:menu nil "\t(&P)"
	   *paste-rectangle-popup-menu*)
	  (:item nil "폜(&D)"
	   'delete-rectangle-selection :modify-rectangle)
	  (:item nil "ر(&R)"
	   'clear-rectangle-selection :modify-rectangle)
	  (:item nil "󔒂}(&O)"
	   'open-rectangle-selection :modify-rectangle)
	  (:item nil "Œu(&S)"
	   'string-rectangle-selection :modify-rectangle)))

  (or *command-bar-popup-menu*
      (setq *command-bar-popup-menu* (create-popup-menu)))

  (or *recent-file-popup-menu*
      (setq *recent-file-popup-menu* (create-popup-menu)))
  (or *recent-session-popup-menu*
      (setq *recent-session-popup-menu* (create-popup-menu)))

  (setq *app-menu*
	(define-menu
	  (:popup 'file "̧(&F)"
	   (:item nil "VK쐬(&N)"
	    'new-file)
	   (:item nil "J(&O)..."
	    'open-file-dialog)
	   (:item nil "(&C)"
	    'close-selected-buffer
	    #'(lambda ()
		(if (= (count-buffers) 1) :disable)))
	   (:item nil "㏑ۑ(&S)"
	    'save-buffer-dialog :modified)
	   (:item nil "Otĕۑ(&A)..."
	    'save-as-dialog)
	   (:item nil "̧ق̑}(&I)..."
	    'insert-file-dialog)
	   (:item nil "O̕ύX(&R)..."
	    'rename-dialog)
	   (:item nil "Sĕۑ(&V)"
	    'save-all-buffers)
	   :sep
	   (:item nil "݂J(&D)..."
	    'open-session-dialog)
	   (:item nil "݂̕ۑ(&W)..."
	    'save-session-dialog)
	   (:item nil "݂̎ۑ(&T)"
	    'toggle-session-auto-save
	    'update-session-auto-save-ui)
	   (:item nil "݂(&Y)"
	    'close-session-dialog)
	   :sep
	   (:item nil "(&P)..."
	    'print-selected-buffer-dialog)
	   :sep
	   (:menu nil "ŋߎģ(&F)"
	    *recent-file-popup-menu*)
	   (:menu nil "ŋߎg(&L)"
	    *recent-session-popup-menu*)
	   :sep
	   (:item nil "ۑďI(&E)"
	    'save-all-buffers-kill-xyzzy)
	   (:item nil "I(&X)"
	    'kill-xyzzy))
	  (:popup 'edit "ҏW(&E)"
	   (:item nil "ɖ߂(&U)"
	    'undo :undo)
	   (:item nil "ς茳ɖ߂Ȃ(&R)"
	    'redo :redo)
	   :sep
	   (:item nil "؂(&T)"
	    'kill-selection-to-clipboard :modify-any-selection)
	   (:item nil "߰(&C)"
	    'copy-selection-to-clipboard :any-selection)
	   (:item nil "\t(&P)"
	    'paste-from-clipboard :clipboard)
	   (:item nil "\t - `(&G)"
	    'paste-rectangle-from-clipboard :clipboard)
	   (:menu nil "دް޴ݺިݸ(&E)"
	    *clipboard-char-encoding-popup-menu*)
	   (:item nil "폜(&L)"
	    'delete-selection :modify-any-selection)
	   (:item nil "SĂI(&S)"
	    'selection-whole-buffer)
	   :sep
	   (:menu nil "`(&N)"
	    *app-rectangle-popup-menu*)
	   (:popup nil "ϊ(&V)"
	    (:item nil "啶(&U)"
	     'upcase-selection :selection)
	    (:item nil "(&L)"
	     'downcase-selection :selection)
	    (:item nil "P̐擪啶(&C)"
	     'capitalize-selection :selection)
	    :sep
	    (:item nil "߰ -> (&T)"
	     'tabify-selection :selection)
	    (:item nil " -> ߰(&S)"
	     'untabify-selection :selection)
	    :sep
	    (:popup nil "Sp -> p(&H)"
	     (:item nil "ׂ(&A)"
	      'map-all-to-half-width-selection :selection)
	     (:item nil "pЂ炪ȥJ^Ji(&S)"
	      'map-to-half-width-selection :selection)
	     (:item nil "pЂ炪(&R)"
	      'map-ascii-and-hiragana-to-half-width-selection :selection)
	     (:item nil "pJ^Ji(&T)"
	      'map-ascii-and-katakana-to-half-width-selection :selection)
	     (:item nil "p(&N)"
	      'map-ascii-to-half-width-selection :selection)
	     (:item nil "Ђ炪(&H)"
	      'map-hiragana-to-half-width-selection :selection)
	     (:item nil "J^Ji(&K)"
	      'map-katakana-to-half-width-selection :selection)
	     (:item nil "MVA(&G)"
	      'map-greek-to-half-width-selection :selection)
	     (:item nil "L(&C)"
	      'map-cyrillic-to-half-width-selection :selection))
	    (:popup nil "p -> Sp(&Z)"
	     (:item nil "ׂ( -> Ђ炪)(&A)"
	      'map-all-to-full-width-hiragana-selection :selection)
	     (:item nil "ׂ( -> J^Ji)(&L)"
	      'map-all-to-full-width-katakana-selection :selection)
	     (:item nil "p( -> Ђ炪)(&R)"
	      'map-to-full-width-hiragana-selection :selection)
	     (:item nil "p( -> J^Ji)(&T)"
	      'map-to-full-width-katakana-selection :selection)
	     (:item nil "p(&N)"
	      'map-ascii-to-full-width-selection :selection)
	     (:item nil " -> Ђ炪(&H)"
	      'map-kana-to-full-width-hiragana-selection :selection)
	     (:item nil " -> J^Ji(&K)"
	      'map-kana-to-full-width-katakana-selection :selection)
	     (:item nil "MVA(&G)"
	      'map-greek-to-full-width-selection :selection)
	     (:item nil "L(&C)"
	      'map-cyrillic-to-full-width-selection :selection))
	    (:popup nil "(&K)"
	     (:item nil "(&A)"
	      'map-char-encoding-selection :selection)
	     (:item nil "JIS(&J)"
	      'map-jis-selection :selection)
	     (:item nil "EUC(&E)"
	      'map-euc-selection :selection)))
	   (:popup nil "i(&H)"
	    (:item nil "O̒i(&B)"
	     'backward-paragraph)
	    (:item nil "̒i(&F)"
	     'forward-paragraph)
	    (:item nil "I(&S)"
	     'selection-paragraph)
	    (:item nil "폜(&L)"
	     'kill-paragraph)
	    (:item nil "ւ(&T)"
	     'transpose-paragraphs)
	    (:item nil "lߍ(&I)"
	     'fill-paragraph))
	   :sep
	   (:item nil "ύXϰ(&M)"
	    'not-modified :modified)
	   (:item nil "ŏ蒼(&O)"
	    'revert-buffer))
	  (:popup 'search "(&S)"
	   (:item nil "(&F)..."
	    'search-dialog)
	   (:item nil "(&N)"
	    'repeat-forward-search
	    #'(lambda () (or *last-search-p* :disable)))
	   (:item nil "O(&P)"
	    'repeat-backward-search
	    #'(lambda () (or *last-search-p* :disable)))
	   (:item nil "u(&R)..."
	    'replace-dialog)
	   (:item nil "v镶𐔂(&C)..."
	    'count-matches-dialog)
	   :sep
	   (:item nil "ŏ(&I)"
	    'first-error)
	   (:item nil "(&E)"
	    'next-error)
	   :sep
	   (:item nil "ws(&J)..."
	    'goto-line-dialog)
	   (:item nil "̧ق̐擪(&T)"
	    'beginning-of-buffer)
	   (:item nil "̧ق̍Ō(&B)"
	    'end-of-buffer)
	   (:item nil "ϰ(&M)..."
	    'mark-dialog-box)
	   (:item nil "֐̐擪(&D)"
	    'beginning-of-defun)
	   (:item nil "֐̍Ō(&U)"
	    'end-of-defun)
	   :sep
	   (:item nil "Gresreg(&S)..."
	    'gresreg-dialog)
	   (:item nil "Grep(&G)..."
	    'grep-dialog))
	  (:popup 'view "\(&V)"
	   (:item nil "܂ԂȂ(&N)"
	    'set-buffer-fold-type-none
	    'set-buffer-fold-type-none-update)
	   (:item nil "wʒuŐ܂Ԃ(&C)"
	    'set-buffer-fold-type-column
	    'set-buffer-fold-type-column-update)
	   (:item nil "޳Ő܂Ԃ(&W)"
	    'set-buffer-fold-type-window
	    'set-buffer-fold-type-window-update)
	   :sep
	   (:menu nil "°ް(&Q)" *command-bar-popup-menu*)
	   :sep
	   (:item nil "Vڰ(&F)"
	    'new-pseudo-frame 'new-pseudo-frame-menu-update)
	   (:item nil "݂ڰт(&D)"
	    'delete-pseudo-frame 'pseudo-frame-menu-update)
	   (:item nil "ڰ(&T)"
	    'other-pseudo-frame 'pseudo-frame-menu-update)
	   (:item nil "Oڰ(&R)"
	    'previous-pseudo-frame 'pseudo-frame-menu-update))
	  (:popup 'window "޳(&W)"
	   (:item nil "ɕ(&S)"
	    'split-window 'split-window-menu-update)
	   (:item nil "cɕ(&V)"
	    'split-window-vertically 'split-window-menu-update)
	   (:item nil "̳޳(&O)"
	    'delete-other-windows 'close-window-menu-update)
	   (:item nil "݂̳޳(&C)"
	    'delete-window 'close-window-menu-update)
	   (:item nil "̳޳(&N)"
	    'other-window 'move-window-menu-update)
	   (:item nil "O̳޳(&P)"
	    'move-previous-window 'move-window-menu-update)
	   :sep
	   (:item nil "xyzzy(&X)"
	    'next-xyzzy-window
	    #'(lambda () (if (< (count-xyzzy-instance) 2) :disable)))
	   (:item nil "Oxyzzy(&Y)"
	    'previous-xyzzy-window
	    #'(lambda () (if (< (count-xyzzy-instance) 2) :disable)))
	   (:item nil "xyzzyI(&Z)..."
	    'select-xyzzy
	    #'(lambda () (if (< (count-xyzzy-instance) 2) :disable)))
	   :sep
	   (:item nil "ޯ̧I(&B)..."
	    'select-buffer))

	  (:popup 'tools "°(&T)"
	   (:item nil "ݿ۸т񓯊Ŏs(&A)..."
	    'execute-subprocess-dialog)
	   (:item nil "񓯊۸тȂ(&K)"
	    'kill-subprocess
	    #'(lambda ()
		(let ((proc (buffer-process (selected-buffer))))
		  (unless (and proc
			       (eq (process-status proc) :run))
		    :disable))))
	   (:item nil "ݿ۸т̎s(&P)..."
	    'pipe-command-dialog)
	   (:item nil "Windows۸т̎s(&W)..."
	    'launch-application-dialog)
	   (:item nil (if (featurep :windows-nt)
			  "NT(&D)"
			"DOS(&D)")
	    'run-console)
	   :sep
	   (:item nil "ްϸۋL^Jn(&S)"
	    'start-kbd-macro)
	   (:item nil "ްϸۋL^I(&E)"
	    'end-kbd-macro
	    :end-macro)
	   (:item nil "ްϸێs(&X)"
	    'call-last-kbd-macro
	    #'(lambda () (if (null *last-kbd-macro*) :disable)))
	   (:popup nil "ްϸۂ܂(&O)"
	    (:item nil "ۑ(&V)..."
	     'save-last-kbd-macro
	     #'(lambda () (if (null *last-kbd-macro*) :disable)))
	    (:item nil "ǂݍ(&L)..."
	     'load-kbd-macro)
	    (:item nil "̧ق֕ۑ(&S)..."
	     'save-kbd-macro-to-file))
	   :sep
	   (:item nil "ʐݒ(&C)..."
	    'option-property-sheet)
	   (:item nil "۰ِݒ(&L)..."
	    'local-option-property-sheet)
	   :sep
	   (:item nil "rӰ(&R)"
	    'box-drawings-mode
	    #'(lambda () (if *box-drawings-mode* :check)))
	   (:item nil "r(&T)"
	    'box-drawings-toggle-line-type
	    #'(lambda ()
		(values (unless *box-drawings-mode*
			  :disable)
			(if (eq *box-drawings-line-type* '*box-drawings-thick-line*)
			    :check))))
	   :sep
	   (:item nil "tƎ̑}(&M)..."
	    'insert-date-string)
	   (:item nil "̧(&F)..."
	    'open-filer)
	   (:menu nil "(&I)"
	    *dictionary-popup-menu*)
	   :sep
	   (:item nil "TAGŞق̍쐬(&G)..."
	    'make-tags-file-dialog)
	   (:item nil "޼ެ(&J)"
	    'jump-tag
	    #'(lambda () (unless (and tags-find-target tags-find-point) :disable)))
	   (:item nil "޲ڸ޼ެ(&U)"
	    'direct-tag-jump
	    #'(lambda () (unless (and tags-find-target tags-find-point) :disable)))
	   (:item nil "ޯ޼ެ(&B)"
	    'back-tag-jump
	    #'(lambda () (if (null *global-mark-list*) :disable)))
	   :sep
	   (:popup nil "(>&_<)"
	    (:item nil "è۸ނǂ(&N)"
	     'view-nifty-log)
	    (:item nil "d(&T)"
	     'calc)
	    (:item nil "ް(&A)"
	     'calendar)
            (:item nil "ɲ̓(&H)"
             'hanoi)
            (:item nil "ܖ(&G)"
	     'gomoku)
            (:item nil "ײ(&L)"
             'life)
            (:item nil "CȐ(&C)"
             'c-curve)
            (:item nil "׺݋Ȑ(&R)"
             'dragon-curve)))
          (:popup 'help "(&?)"
           (:item nil "蓖Ĉꗗ(&K)"
	    'describe-bindings)
	   (:item nil "ŋߓ͂(&L)"
	    'view-lossage)
	   :sep
	   (:item nil "ɂ(&A)..."
	    'about-dialog))))

  (setq *buffer-bar-context-menu*
	(define-popup-menu
	  (:item nil "(&C)"
	   #'(lambda ()
	       (interactive)
	       (kill-buffer *buffer-bar-context-menu-buffer*)))))

  (run-hooks '*init-app-menus-hook*)
  (set-menu *app-menu*)
  (add-file-history-to-menu)
  (add-session-history-to-menu))

(defun make-char-encoding-popup-menu (command &optional init (list *char-encoding-list*))
  (labels ((define-menu (list)
	     (let ((menu (create-popup-menu)))
	       (mapc #'(lambda (encoding)
			 (if (consp encoding)
			     (add-popup-menu menu
					     (define-menu (cdr encoding))
					     (car encoding))
			   (add-menu-item menu nil
					  (char-encoding-display-name encoding)
					  #'(lambda ()
					      (interactive)
					      (funcall command encoding))
					  (and init #'(lambda ()
							(funcall init encoding))))))
		     list)
	       menu)))
    (define-menu
      (if (eq list *char-encoding-list*)
	*char-encoding-tree*
	(labels ((make-tree (tree list)
		   (mapcan #'(lambda (encoding)
			       (if (consp encoding)
				   (let ((l (make-tree (cdr encoding) list)))
				     (and l (list (cons (car encoding) l))))
				 (and (find encoding list :test #'eq)
				      (list encoding))))
			   tree)))
	  (make-tree
	   *char-encoding-tree*
	   (mapcar #'(lambda (encoding)
		       (if (symbolp encoding) (symbol-value encoding) encoding))
	     list)))))))
#|
(defun menu-display-name (path maxlen)
  (when (<= (length path) maxlen)
    (return-from menu-display-name path))
  (let* ((unc )
	 (paths (split-string path #\/))
	 (dev (cond ((eql (string/= path "//") 2)
		     (prog1
			 (concat "//" (car paths) "/" (cadr paths))
		       (setq paths (cddr paths))))
		    (t
		     (pop paths))))
	 (name (progn
		 (setq paths (nreverse paths))
		 (pop paths)))
	 (len (+ (length dev) (length name) 5)))
    (unless name
      (return-from menu-display-name path))
    (when (> len maxlen)
      (return-from menu-display-name name))
    (do* ((npaths nil)
	  (l (length (car paths)) (length (car paths))))
	((> (setq len (+ len l 1)) maxlen)
	 (format nil "~A/.../~{~A/~}~A" dev npaths name))
      (push (pop paths) npaths))))
|#

(defvar *menu-display-length* 40)

(defun add-lru-menu (menu files fn)
  (when (menup menu)
    (while (delete-menu menu 0 t))
    (do ((file files (cdr file))
	 (count 1))
	((or (null file)
	     (> count *menu-history-max*))
	 t)
      (unless (wild-pathname-p (car file))
	(let ((filename (namestring (car file))))
	  (insert-menu-item menu (1- count) nil
			    (format nil "&~:(~35r~): ~a"
				    (rem count 36)
				    (quote-string
				     (abbreviate-display-string
				      filename *menu-display-length* t)
				     #\& #\&))
			    (funcall fn filename)))
	(incf count)))))

(defun add-file-history-to-menu ()
  (when (menup *app-menu*)
    (add-lru-menu *recent-file-popup-menu*
		  *minibuffer-file-name-history*
		  #'(lambda (filename)
		      #'(lambda ()
			  (interactive)
			  (find-file filename)
			  (add-history filename '*minibuffer-file-name-history*))))))
